﻿using System.Collections;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Drawing.Text;

namespace TestProject
{
	/// <summary>
	/// 기본 그리기 객체
	/// </summary>
	public class BaseDrawingObject
	{
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region Field

        /// <summary>
        /// 스무딩 모드
        /// </summary>
        protected SmoothingMode smoothingMode = SmoothingMode.Invalid;

		/// <summary>
		/// 텍스트 렌더링 힌트
		/// </summary>
		protected TextRenderingHint textRenderingHint = TextRenderingHint.SystemDefault;

		/// <summary>
		/// 표시 여부
		/// </summary>
		protected bool isVisible = true;

		/// <summary>
		/// 변환
		/// </summary>
		protected Matrix transform = new Matrix();

		/// <summary>
		/// 부모 그리기 객체
		/// </summary>
		protected IDrawingObject parent = null;

		/// <summary>
		/// 해시 테이블
		/// </summary>
		protected Hashtable hashTable = null;

		/// <summary>
		/// 캐시 그래픽스 경로
		/// </summary>
		protected GraphicsPath cacheGraphicsPath = null;

		/// <summary>
		/// 테두리 내 보이지 않는 객체 포함 여부
		/// </summary>
		protected bool includeInvisibleObjectsInBounds = false;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 부모 그리기 객체 - Parent

        /// <summary>
        /// 부모 그리기 객체
        /// </summary>
        public virtual IDrawingObject Parent
		{
			get
			{
				return this.parent;
			}
		}

        #endregion
        #region 자식 그리기 객체 배열 - ChildArray

        /// <summary>
        /// 자식 그리기 객체 배열
        /// </summary>
		public virtual IDrawingObject[] ChildArray
        {
            get
            {
                return new IDrawingObject[0];
            }
        }

        #endregion

        #region 스무딩 모드 - SmoothingMode

        /// <summary>
        /// 스무딩 모드
        /// </summary>
        public virtual SmoothingMode SmoothingMode
		{
			get
			{
				return this.smoothingMode;
			}
			set
			{
				this.smoothingMode = value;
			}
		}

        #endregion
        #region 텍스트 렌더링 힌트 - TextRenderingHint

        /// <summary>
        /// 텍스트 렌더링 힌트
        /// </summary>
        public virtual TextRenderingHint TextRenderingHint
		{
			get
			{
				return this.textRenderingHint;
			}
			set
			{
				this.textRenderingHint = value;
			}
		}

        #endregion
        #region 표시 여부 - IsVisible

        /// <summary>
        /// 표시 여부
        /// </summary>
        public virtual bool IsVisible
		{
			get
			{
				return this.isVisible;
			}
			set
			{
				this.isVisible = value;
			}
		}

        #endregion

        #region 테두리 내 보이지 않는 객체 포함 여부 - IncludeInvisibleObjectsInBounds

        /// <summary>
        /// 테두리 내 보이지 않는 객체 포함 여부
        /// </summary>
        public bool IncludeInvisibleObjectsInBounds
        {
            get 
            {
                if(!this.includeInvisibleObjectsInBounds && Parent != null)
                {
                    return ((BaseDrawingObject)Parent).IncludeInvisibleObjectsInBounds;
                }

                return includeInvisibleObjectsInBounds;
            }
            set 
            {
                includeInvisibleObjectsInBounds = value;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 자식 찾기 - FindChild(name)

        /// <summary>
        /// 자식 찾기
        /// </summary>
        /// <param name="name">명칭</param>
        /// <returns>그리기 객체</returns>
        public IDrawingObject FindChild(string name)
		{
			return (IDrawingObject)this.hashTable[name.ToLower()];
		}

        #endregion

        #region 변환하기 - Transform()

        /// <summary>
        /// 변환하기
        /// </summary>
        /// <returns>변환 매트릭스</returns>
		public virtual Matrix Transform()
		{
			return Transform(new Matrix());
		}

        #endregion
        #region 변환하기 - Transform(sourceMatrix)

        /// <summary>
        /// 변환하기
        /// </summary>
        /// <param name="sourceMatrix">소스 매트릭스</param>
        /// <returns>변환 매트릭스</returns>
		public virtual Matrix Transform(Matrix sourceMatrix)
        {
            return new Matrix();
        }

        #endregion

        #region 상대적 변환 설정하기 - SetRelativeTransform(sourceMatrix)

        /// <summary>
        /// 상대적 변환 설정하기
        /// </summary>
        /// <param name="sourceMatrix">소스 매트릭스</param>
        public virtual void SetRelativeTransform(Matrix sourceMatrix)
		{
			this.transform = sourceMatrix.Clone();
		}

        #endregion
        #region 상대적 변환 구하기 - GetRelativeTransform()

        /// <summary>
        /// 상대적 변환 구하기
        /// </summary>
        /// <returns>매트릭스</returns>
		public virtual Matrix GetRelativeTransform()
		{
			return this.transform.Clone();
		}

        #endregion
        #region 변환 리셋하기 - ResetTransforms()

        /// <summary>
        /// 변환 리셋하기
        /// </summary>
        public virtual void ResetTransforms()
		{
			this.transform.Reset();

			foreach(IDrawingObject child in ChildArray)
			{
				child.ResetTransforms();
			}
		}

        #endregion
        #region 표준 변환 구하기 - GetStandardTransform(sourceRectangle)

        /// <summary>
        /// 표준 변환 구하기
        /// </summary>
        /// <param name="sourceRectangle">소스 사각형</param>
        /// <returns>표준 변환 매트릭스</returns>
		public virtual Matrix GetStandardTransform(RectangleF sourceRectangle)
		{
			RectangleF boundingRectangle = GetBoundingRectangle();
			RectangleF boundRectangle    = new RectangleF(sourceRectangle.Left, sourceRectangle.Top, sourceRectangle.Width - 1.0f, sourceRectangle.Height - 1.0f);

			float scale = ((boundingRectangle.Width / boundingRectangle.Height < boundRectangle.Width/boundRectangle.Height) ? 1.0f / boundingRectangle.Height * boundRectangle.Height : 1.0f / boundingRectangle.Width * boundRectangle.Width);

			float wo = (boundRectangle.Width  - (boundingRectangle.Width  * scale)) / 2.0f;
			float ho = (boundRectangle.Height - (boundingRectangle.Height * scale)) / 2.0f;

			Matrix matrix = new Matrix();

			matrix.Translate(-boundingRectangle.X, -boundingRectangle.Y, MatrixOrder.Append);
			matrix.Scale(scale, scale, MatrixOrder.Append);
			matrix.Translate(sourceRectangle.Left, sourceRectangle.Top, MatrixOrder.Append);
			matrix.Translate(wo, ho, MatrixOrder.Append);

			return matrix;
		}

        #endregion

        #region 그래픽스 경로 구하기 - GetGraphicsPath()

        /// <summary>
        /// 그래픽스 경로 구하기
        /// </summary>
        public virtual GraphicsPath GetGraphicsPath()
        {
            return GetGraphicsPath(new Matrix());
        }

        #endregion
        #region 그래픽스 경로 구하기 - GetGraphicsPath(sourceMatrix)

        /// <summary>
        /// 그래픽스 경로 구하기
        /// </summary>
        /// <param name="sourceMatrix">소스 매트릭스</param>
        /// <returns>그래픽스 경로</returns>
		public virtual GraphicsPath GetGraphicsPath(Matrix sourceMatrix)
		{
            GraphicsPath graphicsPath = (GraphicsPath)this.cacheGraphicsPath.Clone();

			graphicsPath.Transform(Transform(sourceMatrix));

			return graphicsPath;
		}

        #endregion

        #region 영역 구하기 - GetRegion()

        /// <summary>
        /// 영역 구하기
        /// </summary>
        /// <returns>영역</returns>
        public virtual Region GetRegion()
		{
			return GetRegion(new Matrix());
		}

        #endregion
        #region 영역 구하기 - GetRegion(source)

        /// <summary>
        /// 영역 구하기
        /// </summary>
        /// <param name="source">소스 매트릭스</param>
        /// <returns>영역</returns>
        public virtual Region GetRegion(Matrix source)
		{
			RegionHelper helper = new RegionHelper(GetGraphicsPath(source));

			foreach(BaseDrawingObject child in ChildArray)
			{
				RegionHelper childHelper = child.GetRegionInternally(source);

				helper.Union(childHelper);

				childHelper.Dispose();
			}

			Region resultHelper = helper.GetRegion();

			helper.Dispose();

			return resultHelper;
		}

        #endregion

        #region 바운딩 사각형 구하기 - GetBoundingRectangle()

        /// <summary>
        /// 바운딩 사각형 구하기
        /// </summary>
        public virtual RectangleF GetBoundingRectangle()
        {
            return GetBoundingRectangle(new Matrix());
        }

        #endregion
        #region 바운딩 사각형 구하기 - GetBoundingRectangle(sourceMatrix)       

        /// <summary>
        /// 바운딩 사각형 구하기
        /// </summary>
        /// <param name="sourceMatrix">소스 매트릭스</param>
        /// <returns>바운딩 사각형</returns>
        public virtual RectangleF GetBoundingRectangle(Matrix sourceMatrix)
		{
			RectangleF totalRectangle = new RectangleF();

			SetRuntimeBoundingRectangle(ref totalRectangle, sourceMatrix);

			return totalRectangle;
		}

        #endregion

        #region 중심점 구하기 - GetCenterPoint()

        /// <summary>
        /// 중심점 구하기
        /// </summary>
        /// <returns>중심점</returns>
        public virtual PointF GetCenterPoint()
		{
			return GetCenterPoint(new Matrix());
		}

        #endregion
        #region 중심점 구하기 - GetCenterPoint(sourceMatrix)
        
        /// <summary>
        /// 중심점 구하기
        /// </summary>
        /// <param name="sourceMatrix">소스 매트릭스</param>
        /// <returns>중심점</returns>
		public virtual PointF GetCenterPoint(Matrix sourceMatrix)
		{
			RectangleF boundRectangle = GetBoundingRectangle(sourceMatrix);

			return new PointF(boundRectangle.Width / 2.0f + boundRectangle.Left, boundRectangle.Height / 2.0f + boundRectangle.Top);
		}

        #endregion					

        #region 칠하기 - Paint(graphics)

        /// <summary>
        /// 칠하기
        /// </summary>
        /// <param name="graphics">그래픽스</param>
		public virtual void Paint(Graphics graphics)
		{
			Paint(graphics, GetBoundingRectangle(), 2.0f);
		}

        #endregion
        #region 칠하기 - Paint(graphics, sourceMatrix)

        /// <summary>
        /// 칠하기
        /// </summary>
        /// <param name="graphics">그래픽스</param>
        /// <param name="sourceMatrix">소스 매트릭스</param>
		public virtual void Paint(Graphics graphics, Matrix sourceMatrix)
        {
        }

        #endregion
        #region 칠하기 - Paint(graphics, boundRectangle)

        /// <summary>
        /// 칠하기
        /// </summary>
        /// <param name="graphics">그래픽스</param>
        /// <param name="boundRectangle">테두리 사각형</param>
		public virtual void Paint(Graphics graphics, RectangleF boundRectangle)
		{
			if(GetAutomaticQualityAdjustment(boundRectangle) > 2.0)
			{
				Paint(graphics, boundRectangle, GetAutomaticQualityAdjustment(boundRectangle));
			}
			else
			{
				SmoothingMode smoothingMode = graphics.SmoothingMode;

				graphics.SmoothingMode = SmoothingMode.AntiAlias;

				Paint(graphics, boundRectangle, 1.0f);

				graphics.SmoothingMode = smoothingMode;
			}
		}

        #endregion
        #region 칠하기 - Paint(graphics, boundRectangle, quality)

        /// <summary>
        /// 칠하기
        /// </summary>
        /// <param name="graphics">그래픽스</param>
        /// <param name="boundRectangle">테두리 사각형</param>
        /// <param name="quality">품질</param>
		public virtual void Paint(Graphics graphics, RectangleF boundRectangle, float quality)
		{
			Matrix matrix = new Matrix();

			Paint(graphics, boundRectangle, quality, out matrix);
		}

        #endregion
        #region 칠하기 - Paint(graphics, boundRectangle, quality, targetMatrix)

        /// <summary>
        /// 칠하기
        /// </summary>
        /// <param name="graphics">그래픽스</param>
        /// <param name="boundRectangle">테두리 사각형</param>
        /// <param name="quality">품질</param>
        /// <param name="targetMatrix">타겟 매트릭스</param>
		public virtual void Paint(Graphics graphics, RectangleF boundRectangle, float quality, out Matrix targetMatrix)
		{
			targetMatrix = new Matrix();

			if(boundRectangle.Width <= 2.0 || boundRectangle.Height <= 2.0)
            {
                return;
            }

			if(quality > 1.0)
			{
				Paint(graphics, boundRectangle, quality, out targetMatrix, null);
			}
			else
			{
				targetMatrix = GetStandardTransform(boundRectangle);

				Matrix saveMartix = graphics.Transform;

				graphics.Transform = new Matrix();

				Paint(graphics, targetMatrix);

				graphics.Transform = saveMartix;
			}
		}

        #endregion
        #region 칠하기 - Paint(graphics, boundRectangle, quality, targetMatrix, imageAttribute)

        /// <summary>
        /// 칠하기
        /// </summary>
        /// <param name="graphics">그래픽스</param>
        /// <param name="boundRectangle">테두리 사각형</param>
        /// <param name="quality">품질</param>
        /// <param name="targetMatrix">타겟 매트릭스</param>
        /// <param name="imageAttribute">이미지 어트리뷰트</param>
		public virtual void Paint(Graphics graphics, RectangleF boundRectangle, float quality, out Matrix targetMatrix,  ImageAttributes imageAttribute)
		{
			bool useGDIPlusAntiAlias = (graphics.SmoothingMode == SmoothingMode.AntiAlias || graphics.SmoothingMode == SmoothingMode.HighQuality);

			if(useGDIPlusAntiAlias)
			{
				targetMatrix = GetStandardTransform(boundRectangle);

				Matrix saveMatrix = graphics.Transform;

				graphics.Transform = new Matrix();

				Matrix cloneMatrix = targetMatrix.Clone();

				cloneMatrix.Multiply(saveMatrix);

				Paint(graphics, cloneMatrix);

				graphics.Transform = saveMatrix;
			}
			else
			{
				float actualQuality = (quality < 1.0f ? 1.0f : (quality > 8.0f ? 8.0f : quality));

				targetMatrix = GetStandardTransform(boundRectangle);

				if(boundRectangle.Width <= 0 || boundRectangle.Height <= 0)
                {
                    return;
                }

				Matrix saveMatrix = graphics.Transform;

				graphics.Transform = new Matrix();

				Bitmap bitmap = new Bitmap((int)(boundRectangle.Width * actualQuality), (int)(boundRectangle.Height * actualQuality));

				Graphics bitmapGraphics = Graphics.FromImage(bitmap);

				Matrix cloneMatrix = targetMatrix.Clone();

				cloneMatrix.Multiply(saveMatrix);

				cloneMatrix.Translate(-boundRectangle.Left, -boundRectangle.Top, MatrixOrder.Append);

				cloneMatrix.Scale(actualQuality, actualQuality, MatrixOrder.Append);

				Paint(bitmapGraphics, cloneMatrix);

				graphics.InterpolationMode = InterpolationMode.HighQualityBicubic;

				if(imageAttribute != null)
				{
					graphics.DrawImage(bitmap, Rectangle.Round(boundRectangle), 0, 0, bitmap.Width, bitmap.Height, GraphicsUnit.Pixel, imageAttribute);
				}
				else
				{
					graphics.DrawImage(bitmap, Rectangle.Round(boundRectangle), 0, 0, bitmap.Width, bitmap.Height, GraphicsUnit.Pixel);

				}

				bitmapGraphics.Dispose();

				bitmap.Dispose();

				graphics.Transform = saveMatrix;
			}
		}

        #endregion

        #region 모든 바운딩 사각형 구하기 - GetAllBoundingRectangle(source)
        
        /// <summary>
        /// 모든 바운딩 사각형 구하기
        /// </summary>
        /// <param name="source">소스 매트릭스</param>
        /// <returns>모든 바운딩 사각형</returns>
        public virtual RectangleF GetAllBoundingRectangle(Matrix source)
		{
			bool isSaved = this.includeInvisibleObjectsInBounds;

			this.includeInvisibleObjectsInBounds = true;

			RectangleF totalRectangle = new RectangleF();

			SetRuntimeBoundingRectangle(ref totalRectangle, source);

			this.includeInvisibleObjectsInBounds = isSaved;

			return totalRectangle;
		}

        #endregion
        #region 모든 바운딩 사각형 구하기 - GetAllBoundingRectangle()

        /// <summary>
        /// 모든 바운딩 사각형 구하기
        /// </summary>
        /// <returns>모든 바운딩 사각형</returns>
        public virtual RectangleF GetAllBoundingRectangle()
        {
            return GetAllBoundingRectangle(new Matrix());
        }

        #endregion

        #region 자동 품질 조정 구하기 - GetAutomaticQualityAdjustment(boundRectangle)

        /// <summary>
        /// 자동 품질 조정 구하기
        /// </summary>
        /// <param name="boundRectangle">테두리 사각형</param>
        /// <returns>자동 품질 조정</returns>
		public virtual float GetAutomaticQualityAdjustment(RectangleF boundRectangle)
		{
			float size = (float)System.Math.Log(boundRectangle.Width * boundRectangle.Height);

			if(size > 14.0f)
            {
                return 2.0f;
            }

			if(size > 7.0f)
            {
                return 4.0f;
            }

			return 8.0f;
		}

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected
					
        #region 런타임 바운딩 사각형 설정하기 - SetRuntimeBoundingRectangle(totalRectangle, sourceMatrix)

        /// <summary>
        /// 런타임 바운딩 사각형 설정하기
        /// </summary>
        /// <param name="totalRectangle">전체 사각형</param>
        /// <param name="sourceMatrix">소스 매트릭스</param>
		protected virtual void SetRuntimeBoundingRectangle(ref RectangleF totalRectangle, Matrix sourceMatrix)
		{
			if(!IsVisible) 
			{
				if(!IncludeInvisibleObjectsInBounds) 
				{
					return;
				}
			}
								
			RectangleF temporaryRectangle = GetGraphicsPath(sourceMatrix).GetBounds();

			if(!temporaryRectangle.IsEmpty)
			{
				if(totalRectangle.IsEmpty)
                {
					totalRectangle = temporaryRectangle;
                }
				else
                {
					totalRectangle = RectangleF.Union(totalRectangle, temporaryRectangle);
                }
			}

			foreach(BaseDrawingObject child in ChildArray)
			{
				child.SetRuntimeBoundingRectangle(ref totalRectangle, sourceMatrix);
			}
		}

        #endregion
        #region 영역 구하기 (내부용) - GetRegionInternally(source)

        /// <summary>
        /// 영역 구하기 (내부용)
        /// </summary>
        /// <param name="source">소스 매트릭스</param>
        /// <returns>영역 헬퍼</returns>
        protected virtual RegionHelper GetRegionInternally(Matrix source)
		{
			RegionHelper helper = new RegionHelper(GetGraphicsPath(source));

			foreach(BaseDrawingObject child in ChildArray)
			{
				RegionHelper childHelper = child.GetRegionInternally(source);

				helper.Union(childHelper);

				childHelper.Dispose();
			}

			return helper;
		}

        #endregion
	}
}